# sparksdr2all.py (2024-05-01) by AndyZ K1RA https://k1ra.us/
#
# Usage: sparksdr2all.py -f MHz[,MHz,...] [-s sparksdr_ip] [-d] [-v] [-h]
#
# Create a WSJT-X ALL.txt file in the standard WSJT-X data directory ex: C:\Users\k1ra\AppData\Local\WSJT-X\ALL.TXT
# for programs like SPOTS to utilize, pulling spots from SparkSDR https://sparksdr.com/ Websocket data for FT8, FT4,
# WSPR, FST4W, JS8 decodes on all bands
#
#
# use "pip install -r requirements.txt" to install python dependencies and libraries needed by this app

import json
import os
import re
import time
import threading
import sys
import getopt
import psutil as psutil

from websocket import WebSocketApp
from multiprocessing import Process, Queue
from threading import Thread
from queue import Queue
from functools import partial

version = "2.0.4"


# Create a Timer class for scheduling tasks (currently not used)
class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time.time()
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time.time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


# Create a Websocket client class for connecting to SparkSDR server
class Websocket_Client:
    def __init__(self, wsurl: str):
        self.responseQ = Queue()  # our message response Q

        # add the response Qs on by wrapping the callbacks in functools partials
        self.websocket = WebSocketApp(wsurl,
                                      on_open=partial(self.on_open, responseQ=self.responseQ),
                                      on_message=partial(self.on_message, responseQ=self.responseQ),
                                      on_error=partial(self.on_error, responseQ=self.responseQ),
                                      on_close=partial(self.on_close, responseQ=self.responseQ))

        # start run_forever as thread. request keep alive pings every 10 seconds with 5 second timeouts
        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    # open connection and send subscribe message to begin receiving spots
    def on_open(self, websocket, responseQ):
        global debug

        if debug:
            print("Opening Connection")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")

    def on_error(self, websocket, error, responseQ):
        global debug

        if debug:
            print("Connection Error:", error)

    # get a SparkSDR spot decode message and copy to the queue for processing
    def on_message(self, websocket, message, responseQ):
        global debug

        if debug:
            print("Message Received:", message)
        responseQ.put(message)

    # connection closed and stop receiving sport, but spawn reconnect attempts every 2 seconds
    def on_close(self, websocket, status, message, responseQ):
        global debug

        if debug:
            print("Closing Connection:")
        time.sleep(2)

        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    # force a connection close
    def stop(self):
        self.websocket.keep_running = False
        self.websocket.close()


# Parse a SparkSDR Websocket JSON message into individual fields
def parse_message(json_message):
    global debug

    date = json_message["time"]

    # parse time
    m = re.search('\d{2}(\d{2})-(\d{2})-(\d{2})T', date)
    YY = m.group(1)
    MM = m.group(2)
    DD = m.group(3)
    YYMMDD = f"{YY}{MM}{DD}"

    m = re.search('(\d{2}):(\d{2}):(\d{2})', date)
    hh = m.group(1)
    mm = m.group(2)
    ss = m.group(3)
    hhmmss = f"{hh}{mm}{ss}"

    timedate = YYMMDD + "_" + hhmmss

    call = json_message["call"]
    mode = json_message["mode"]

    freq = json_message["tunedfrequency"]
    freq = freq / 1000.

    tone = int(json_message["frequency"]) - int(json_message["tunedfrequency"])
    dist = json_message["distance"]
    snr = str(json_message["snr"])
    loc = json_message["locator"]

    # parse grid
    if loc == "":
        grid = "    "
    else:
        m = re.search('([A-Z][A-Z]\d\d)', loc)
        grid = m.group(1)

    dt = float(json_message["dt"])
    # valid = json_message["valid"]
    valid = ""
    jtmsg = json_message["msg"]

    jtmsg = re.sub('[\r\n]', '', jtmsg)

    called = ""

    # handle FTn messages
    if re.search('FT\d', mode) is not None:
        # look for different FTn message structures and extract calls and grid

        # CQ CALL
        if (m := re.search('^CQ\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)

        # CQ CALL LLnn
        elif (m := re.search('^CQ\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # [NA, DX, SD, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)
            grid = ""

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 [R][-+]nn
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*[\-+][0-9]{2}$',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 RRR
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sRRR$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 GRID
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\s([A-Z][A-Z]\d\d$)',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL GRID
        elif (m := re.search('^\s*([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d)$', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2 RR73 or 73
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*73$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # SP9VEY RR73; IK6CLX <TY0RU> -08
        elif (
                m := re.search('^[A-Z\d\/]{3,}\sRR73;\s([A-Z\d\/]{3,})\s\<(\.\.\.|[A-Z\d\/]{3,})\>\s',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 R GRID
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\sR\s([A-Z][A-Z]\d\d$)',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL1 CALL2 RST nnnn
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\s\d{3}\s\d{4}$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""
        # no match
        else:
            if debug:
                if call != "":
                    print(f"<<<--- NO MATCH --->>> {call}, {jtmsg}")
            return False, timedate, freq, mode, call, grid, tone, snr, dt, "", jtmsg

        if re.search('RR73', grid) is not None:
            grid = ""

    # pull fields only found in WSPR messages
    if mode == "WSPR" or mode == "FST4W":
        # WSPR
        # watt = json_message["watts"]
        dbm = json_message["power"]
        jtmsg = f"CQ {call} {grid}"

        if re.search('[A-Z][A-Z]\d\d', grid) is None:
            return False, timedate, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg
    else:
        watt = ""
        dbm = ""

    call = re.sub('[<>\.]', '', call)

    # perform some sanity checks on calls, check for bogus combos and return False if necessary
    if (call == "" or re.search('RR73', call) is not None) or (re.search('\d[A-Z]', call) is None) or \
            (re.search('^\d{2;}[A-Z]{2,}', call) is not None) or (re.search('^[0Q]', call) is not None) or \
            (re.search('^[0Q]', call) is not None) or (re.search('^\d.\d.*\/[PQR]$', call) is not None) or \
            (re.search('^\d{2;}[A-Z]{2,}', called) is not None) or (re.search('_|\/.*\/]', call) is not None) or \
            (re.search('\d[A-Z]+\d[A-Z]+\d', call) is not None) or \
            ((re.search('\/[PQR]$', call) is not None) and (re.search('\/[PQR]$', called) is not None)) or \
            ((re.search('K1ABC', called) is not None) and (re.search('\/[PQR]$', call) is not None)):
        return False, timedate, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg

    # get current time
    nowtime = int(time.time())

    return valid, timedate, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg


# Start sparksdr2all.py
if __name__ == '__main__':
    debug = False
    sparksdr_ip = "127.0.0.1"
    sparksdr_port = 4649
    filtermhz = ""

    # using path.expanduser() getting username
    ALLTXT = os.path.expanduser('~') + "\\AppData\\Local\\WSJT-X\\ALL.TXT"

    try:
        opts, args = getopt.getopt(sys.argv[1:], "f:s:dvh")
    except getopt.GetoptError:
        print("sparksdr2all.py -f MHz[,MHz,...] [-s sparksdr_ip] [-d] [-v] [-h]")
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print("sparksdr2all.py -f MHz[,MHz,...] [-s sparksdr_ip] [-d] [-v] [-h]")
            print("-f MHz (mandatory: MHz bands, ex: -f 24,28)")
            print("-s sparksdr_ip (optional: ip address, ex: -s 192.168.1.123) [default=127.0.0.1]")
            print("-d (turns on debug messages)")
            print("-v (prints version)")
            print("-h (prints this help)")
            sys.exit()
        elif opt in "-f":
            filtermhz = arg.split(",")
        elif opt in "-d":
            debug = True
        elif opt in "-s":
            sparksdr_ip = arg
            if re.search('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', sparksdr_ip) is None:
                print(f"sparksdr_ip address {sparksdr_ip} incorrect")
                sys.exit(3)
        elif opt in "-v":
            print(f"sparksdr2all version {version} by AndyZ K1RA")
            sys.exit()

    if filtermhz == "":
        print("sparksdr2all.py -f MHz[,MHz,...] [-s sparksdr_ip] [-d] [-v] [-h]")
        sys.exit(2)

    print(f"sparkdxc2all version {version} by K1RA")

    url = f"ws://{sparksdr_ip}:{sparksdr_port}/Spark"

    # make the websocket client
    wsclient = Websocket_Client(url)

    pid = os.getpid()

    try:
        # do forever
        while True:
            # process FT8, FT4, WSPR websocket messages from SparkSDR queue
            while not wsclient.responseQ.empty():
                spark_msgs = wsclient.responseQ.get()

                json_msgs = json.loads(spark_msgs)

                printed = False

                # for every individual decode spot
                if "spots" in json_msgs:
                    for json_msg in json_msgs["spots"]:

                        valid, timedate, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg = parse_message(json_msg)

                        # don't process or report call if it was recently seen within expire timeout
                        if valid == "FRESH" or call == "":
                            continue

                        if debug and (valid is False or valid == "false"):
                            print()
                            print("<<<--- INVALID --->>> {}  {:7.1f} {:>4} {:>11} {:>4} {:>3} {:>4}  {}".format(
                                timedate, freq, mode, call, tone, snr, grid, jtmsg))
                            print()
                            continue

                        mhz = str(int(freq / 1000.))

                        if mhz in filtermhz:
                            # spot = sprintf "%s %9.3f Rx %-4s   %3s  0.0 %4s %s\n",
                            # $time, $freq / 1000., $mode, $snr, $hz, $jtmsg;

                            original_stdout = sys.stdout  # Save a reference to the original standard output

                            freq = freq / 1000.

                            # 210102_182345    14.074 Rx FT8      1  0.2 1380 KA0BOJ 5Z4VJ RR73
                            with open(ALLTXT, 'a') as f:
                                sys.stdout = f  # Change the standard output to the file we created.
                                print(
                                    "{} {:9.3f} Rx {:<4}   {:>3} {:>4} {:>4} {}".format(timedate, freq, mode, snr,
                                                                                        dt, tone, jtmsg))
                                sys.stdout = original_stdout  # Reset the standard output to its original value

                            if debug:
                                print(
                                    "{} {:9.3f} Rx {:<4}   {:>3} {:>4} {:>4} {}".format(timedate, freq, mode, snr,
                                                                                        dt, tone, jtmsg))
                                printed = True

                    if debug and printed:
                        print()

            time.sleep(1)

            now_time = int(time.time())

    except KeyboardInterrupt:
        if debug:
            print("Closing main thread")

        p = psutil.Process(pid)
        p.terminate()  # or p.kill()

        wsclient.controller.join()
